const config=new Object();

function clear(){
	let el=config.document.getElementById(config.debugId);
	if (el===null) {
		//alert("Debug section not initialized");
		return;
	}
	config.document.getElementById(config.debugId).value="";
}

function print(msg){
	let el=config.document.getElementById(config.debugId);
	if (el===null) {
		//alert("Debug section not initialized. Message: "+msg);
		return;
	}
	el.value+="\n"+msg;
}

//harmonize signature
//(schema,query,row,action=cvDoNothing,container=null


function cvInsertRows(schema,query,rows,action=null,container=null,failure=null){
   cvModifyRows('insert',schema,query,rows,action,container);
}


function cvDeleteRows(schema,query,rows,action=null,container=null,failure=null){
   cvModifyRows('delete',schema,query,rows,action,container);
}

function cvModifyRows(mode,schema,query,rows,action=null,container=null,failure=null){
	//insert rows to container/schema/query and return with action
	let fName="[cvModifyRows/"+mode+"]";
	print(fName+' '+schema+'/'+query);
	let qconfig=new Object();
	qconfig.schemaName=schema;
	qconfig.queryName=query;
   if (container) qconfig.containerPath=container;
   if (!rows) {
      print(fName+' rows '+rows);
      return;
   }
	qconfig.rows=rows;
   qconfig.success=function(data){;};
	if (action) qconfig.success=action;
   if (mode=='insert') LABKEY.Query.insertRows(qconfig);
   if (mode=='update') LABKEY.Query.updateRows(qconfig);
   if (mode=='delete') LABKEY.Query.deleteRows(qconfig);
	print(fName+" done");
}

function cvSelectRows(schema,query,filters=[],action=null, container=null, failure=null, columns=null){
	let fName="[cvSelectRows]";
	print(fName+' '+schema+' '+query+' '+container);
	let qconfig=new Object();
	qconfig.schemaName=schema;
	qconfig.queryName=query;
   if (container) qconfig.containerPath=container;
   qconfig.filterArray=filters;
   qconfig.success=function(data){;};
	if (action) qconfig.success=action;
   if (failure) qconfig.failure=failure;
   if (columns) qconfig.columns=columns;
	LABKEY.Query.selectRows(qconfig);
	print(fName+" done");

}

function createCrfStatus(crfEntry){
   let crfStatus=new Object();
   crfStatus.entryId=crfEntry.entryId;
   crfStatus.submissionDate=new Date();
   crfStatus.FormStatus=crfEntry.FormStatus;
   crfStatus.User=crfEntry.UserId;
   crfStatus.Form=crfEntry.Form;
   return crfStatus;
}


function getCRFrefFirst(){
	//crfRef is part of html call and gets stored in the page
	return config.document.getElementById(config.crfRefId).innerHTML;
}


function getCRFref(){
	//'crfRefId'
	return	config.formConfig.crfEntry['entryId'];
}

function getCRFrefData(){
	let parentCrf=config.formConfig.crfEntry['parentCrf'];
	if (parentCrf!=undefined) return parentCrf;
	return getCRFref();
}


function onFailure(errorInfo, options, responseObj){
	
	if (errorInfo && errorInfo.exception)
		alert("Failure: " + errorInfo.exception);
	else
		alert("Failure: " + responseObj.statusText);
}

function doNothing(){
	print('doNothing called');
}


function createQueryObject(){
   //create object ready for makeQuery
   let x=new Object();
   x.print=print;
   x.getContainer=getContainer;
   return x;
}

function getSnapshotObject(){
   if (!("dataQueriesSnapshot" in config.formConfig))
      config.formConfig.dataQueriesSnapshot=createQueryObject();
   return config.formConfig.dataQueriesSnapshot;
}

function getQuerySnapshot(queryName){
   //check whether queryName is in snapshotObject?
   return getSnapshotObject()[queryName];
}

function getLayoutObject(){
   if (!("dataQueriesLayout" in config.formConfig))
      config.formConfig.dataQueriesLayout=createQueryObject();
   return config.formConfig.dataQueriesLayout;
}

function getQueryLayout(queryName){
   //check whether queryName is in snapshotObject?
   return getLayoutObject()[queryName];
}

function getLookupObject(){
   if (!("lookup" in config.formConfig))
      config.formConfig.lookup=createQueryObject();
   return config.formConfig.lookup;
}

function getLookup(queryName){
   let x=getLookupObject();
   if (queryName in x) return x[queryName];
   return null;
}

function getQueryList(){
   if (!("queryList" in config.formConfig))
      config.formConfig.queryList=new Object();
   return config.formConfig.queryList;
}

function generateDebugSection(){
	//let debug=true;
	//if (debug) print("generateDebugSection "+sectionName);

	let formName=config.debugDiv;
	let sectionName="debugSection";
	let sectionTitle="Debug Messages";
	let tb=config.document.createElement('table');
	tb.className='t2';
	let row=tb.insertRow();
	let cell=config.document.createElement('th');
	row.appendChild(cell);	
	cell.setAttribute("colspan","4");
	cell.style.fontSize="20px";
	cell.style.textAlign="center";
	let cellData=config.document.createTextNode(sectionTitle);
	cell.appendChild(cellData);
	cell=row.insertCell();
	let input=config.document.createElement("input");	
	input.type="button";
	input.id="toggle"+sectionName+"VisbilityButton";
	input.onclick=function(){toggleVisibility(sectionName,input.id)};
	cell.appendChild(input);
	config.document.getElementById(formName).appendChild(tb);
	

	let div=config.document.createElement('div');
	div.id=sectionName;
	config.document.getElementById(formName).appendChild(div);	
	
	//start open (for debug)
	
	//input.value="Hide";
	//div.style.display="block";

	//start hidden (for production)
	input.value="Show";
	div.style.display="none";

	let debugArea=config.document.createElement('textarea');
	debugArea.rows=10;
	debugArea.cols=95;
	debugArea.id=config.debugId;
	div.appendChild(debugArea);

	//print('ver: 0.10');

}




function getAdditionalData(formSetupEntry){
   //return information on additional data associated with the form
   //additionalData is a sub-list with multiple entries per patient/visit
   

   //argument is the row of the formSetup setup list
	let queryName=config.formConfig.queryMap[formSetupEntry['queryName']];
	let fName='[getAdditionalData/'+queryName+']';
	print(fName);

   //additionalData holds a reference to all queries already parsed
   //this helps in reducing number of calls to the database (I assume)
	if (queryName in config.formConfig.additionalData){
		print(fName+': Returning preset value');
		return config.formConfig.additionalData[queryName];
	}

   //first time we see this query, so we have to do the setup
	print(fName+': generating');
	config.formConfig.additionalData[queryName]=new Object();
	
   //takes address, so further changes will be to the newly created object
   //in fact, ad is just a short alias of the long variable name on the right
	let ad=config.formConfig.additionalData[queryName];

   //no additional data
	if (formSetupEntry["showFlag"]==="NONE") {
		print(fName+": empty");
		return ad;
	}

   //use showFlag to setup report section of the CRF list
	if (formSetupEntry["showFlag"]==="REVIEW") {
		//abuse additionalData to signal different segment
		print(fName+": generateReport");
		ad.isReview=true;
		return ad;
	}

   //setup the additionalData memory object
	print(fName+': setting values');
	ad.showFlag=formSetupEntry["showFlag"];
	ad.showFlagValue=formSetupEntry["showFlagValue"];
	ad.queryName=formSetupEntry["showQuery"];

   //for data queries, limit to present CRF only
	ad.filters=new Object();
	ad.filters['crfRef']=getCRFref();

   //compose a long debug message
	let msg=fName+": flag "+ad.showFlag;
	msg+=" value "+ad.showFlagValue;
	msg+=" query "+ad.queryName;
	print(msg);

	return ad;	
}

function fullAccessSetup(listName){
	//generate setup object whcih should contain fields:
	//readonlyFlag - whether the dataset is writeable
	//filters - selection fields that allow creation of LABKEY.Filter.create()
	//getInputId - formating of unique ids for html elements
	//addApply - whether a submit/Save button is generated
	//unique - whether entries in list are unique

	let debug=true;
	if (debug) print("fullAccessSetup");

	let setup=new Object();
	setup.queryName=listName;
	setup.readonlyFlag=function(vName){return false};
	setup.filters=new Object();
	setup.filters['crfRef']=getCRFref();
	setup.getInputId=function(vName){return listName+"_"+vName;}
	setup.addApply="Save";
	setup.isReview=false;
	return setup;

}

function readonlySetup(listName){
	//see definition of setup object above
	let debug=true;
	if (debug) print("readonlySetup");

	let setup=new Object();
	setup.queryName=listName;
	setup.readonlyFlag=function(vName){return true};
	setup.filters=new Object();
	setup.filters['crfRef']=getCRFref();
	setup.getInputId=function(vName){return listName+'_'+vName;}
	setup.isReview=false;
	return setup;
}

function getSetup(listName,writeAccess=true){
	//change to section granulated permission of type EDIT, COMMENT, READ
	//let formStatus=config.formConfig.formStatus;
	//equivalent to READ

	if (!writeAccess)
	//if (formStatus=="Submitted")
		return readonlySetup(listName);
	//if (formStatus=="Approved")
	//	return readonlySetup(listName);
	return fullAccessSetup(listName);
}

function generateSection(formSetupEntry){

	let listName=config.formConfig.queryMap[formSetupEntry['queryName']];
	let fName='[generateSection/'+listName+']';
	let sectionTitle=formSetupEntry['title'];	
	let accessModeColumn=config.formConfig.operator+'Mode';
	let accessMode=formSetupEntry[accessModeColumn];
	//this will fix it for later use as well
	let additionalData=getAdditionalData(formSetupEntry);
	print(fName);

	let formName=config.masterForm;//this is HTML designator of area on page
	let debug=true;
	let tb=config.document.createElement('table');
	tb.className='t2';
	let row=tb.insertRow();
	let cell=config.document.createElement('th');
	row.appendChild(cell);	
	cell.setAttribute("colspan","4");
	cell.style.fontSize="20px";
	cell.style.textAlign="center";
	let cellData=config.document.createTextNode(sectionTitle);
	cell.appendChild(cellData);
	cell=row.insertCell();
	let input=config.document.createElement("input");	
	input.type="button";
	input.value="Show";
	input.id="toggle"+listName+"VisbilityButton";
	input.onclick=function(){toggleVisibility(listName,input.id)};
	cell.appendChild(input);
	config.document.getElementById(formName).appendChild(tb);

	let div=config.document.createElement('div');
	div.id=listName;
	div.style.display="none";
	config.document.getElementById(formName).appendChild(div);	

	let divTable=config.document.createElement('div');
	divTable.id=listName+"Table";
	div.appendChild(divTable);
	
	if ("showFlag" in additionalData) {

		additionalData.divName=listName+"SubDiv";
		additionalData.divQueryName=listName+"SubDivList";

		let div1=config.document.createElement('div');
		div1.id=additionalData.divName;
		div1.style.display="none";
		div.appendChild(div1);
	
		let div2=config.document.createElement('div');
		div2.id=additionalData.divQueryName;
		div1.appendChild(div2);

	}
	if (debug) print("generate master table");

	let writeMode=accessMode=="EDIT";	
	let setup=getSetup(listName,writeMode);
	
	if ("isReview" in additionalData){
		generateReviewSection(listName,div.id,generateReviewSectionCB);
		return;	
	}
	//master table is unique per visit

	
	setup.unique=true;
	generateTable(listName,divTable.id,additionalData,setup);
	
	if (debug) print("generate master table: done");

	let generateSubTable=true;
	//generateSubTable equivalent to read/write access to section
	if (accessMode != "EDIT")
		generateSubTable=false;
	
	if (! ("showFlag" in additionalData) ) generateSubTable=false;
	
	if (generateSubTable){
		let qName=additionalData.queryName;
		let dName=additionalData.divName;
		

		let xsetup=fullAccessSetup(qName);
		//only set master query for additionalData
		xsetup.masterQuery=listName;
		//if (readonly) setup=readonlySetup(config);
      xsetup.subTable=true;
		generateTable(qName,dName,additionalData,xsetup);
		//generateTable(formSetupEntry,qName,dName,additionalData,setup);
	}

	print("generate review");

	let divReviewList=config.document.createElement('div');
	divReviewList.id=listName+"ReviewList";
	div.appendChild(divReviewList);
	
	let divReview=config.document.createElement('div');
	divReview.id=listName+"Review";
	div.appendChild(divReview);


	//assume we already have listId (content of config.setupQueryName is listId)
	//we need listName also
	//qconfig.queryName=config.setupQueryName;
	generateReview(divReview.id,divReviewList.id,listName,accessMode);

	if (accessMode!='GENERATE') return;
	print('Adding generate button');	
	//add generateButton
	let divGenerateButton=config.document.createElement('div');
	divGenerateButton.id=listName+'GenerateButton';
	div.appendChild(divGenerateButton);
	print('Adding generate button completed to here');	
	let cb=function(){onGenerateQuery(listName);};
	generateButton(divGenerateButton.id,'Generate','Generate '+listName,'XX',cb);
	print('Adding generate button completed');	
}

//>>>>reviewSection associated routines

function parseResponseXML(){
	//print(config.config,'Status:' +this.status);
	print('Status:'+this.status);
	if (this.status!=200) return;
	config.loadFileConfig.json=JSON.parse(this.responseText);
	config.loadFileConfig.cb();
}

function loadFile(){
	print('YY: '+config.loadFileConfig.url);

	let connRequest=new XMLHttpRequest();
	connRequest.addEventListener("loadend",parseResponseXML);
		//function(e){parseResponseXML(e,config);});
	connRequest.open("GET", config.loadFileConfig.url);
	connRequest.send();
}


function getBasePath(){
	let server=LABKEY.ActionURL.getBaseURL();
	let basePath=server+"_webdav";
	basePath+=LABKEY.ActionURL.getContainer();
	return basePath;
}


function generateErrorMessage(id,listName,msg){
	print('generateErrorMessage:');
	let eid=listName+"_errorMsg";
	let el=config.document.getElementById(eid);
	if (el===null){
		el=config.document.createElement("p");
		config.document.getElementById(id).appendChild(el);
	}
	el.innerHTML=msg;
}

function clearErrorMessage(listName){
	let eid=listName+"_errorMsg";
	let el=config.document.getElementById(eid);
	if (el===null) return;
	el.remove();
}



function generateReview(divReviewId,divReviewListId, listName, accessMode){
	let listId=config.formConfig.fields[listName].queryId;

	//listId is a number->should it be queryName?
	

	let debug=true;
	if (debug) print("Generate review for: "+listId+'/'+listName);
	
	let reviewSetup=new Object();
	reviewSetup.readonlyFlag=function(vName){
		if (vName=="queryName") return true; 
		if (vName=="queryname") return true; 
		if (vName=="ModifiedBy") return true;
		return false;};
	reviewSetup.addApply="Add Review";
   reviewSetup.reviewTable=true;

	let generateTableFlag=true;
	let formStatus=config.formConfig.formStatus;
	//COMMENTS allowed or not
	//three levels of access: EDIT, COMMENT, READ
	if (accessMode == "READ"){
	//if (formStatus == "Approved" ){
		delete reviewSetup.addApply;
		reviewSetup.readonlyFlag=function(vName){return false;}
		generateTableFlag=false;
	}
	
	reviewSetup.filters=new Object();
	reviewSetup.filters["crfRef"]=getCRFref();
   if (config.formConfig.crfEntry.parentCrf){
      reviewSetup.filters["crfRef"]=getCRFref()+";"+config.formConfig.crfEntry.parentCrf;
   }
 	reviewSetup.filters["queryName"]=listId;//entry in reviewComments list is queryname, all in small caps
	//needs listName, in argument
	
	reviewSetup.getInputId=function(vName){return listName+"_add"+vName};
	reviewSetup.divReviewListId=divReviewListId;
	reviewSetup.isReview=true;	

	if (debug) {
		let msg="Review: divId: "+divReviewId;
		msg+=" inputId: "+reviewSetup.getInputId;
		print(msg);
	}
	updateListDisplay(divReviewListId,"reviewComments",reviewSetup.filters,true);

	if (! generateTableFlag) return;

	
	generateTable("reviewComments",divReviewId,new Object(),reviewSetup);
}	

//>>>>>>>>>>trigger visibility of additional lists

function setListVisibility(input,setup,readonlyFlag){
	let debug=true;
	let fName="[setListVisibility/"+setup.queryName+"]";
	print(fName);
	let additionalData=config.formConfig.additionalData[setup.queryName];
	
	let x = config.document.getElementById(additionalData.divName);
	if (debug) print(fName+": Div: "+x);
	x.style.display="none";

	let sText;
	if (readonlyFlag) sText=input.innerText;
	else sText=input.options[input.selectedIndex].text;
			
	if (debug) print(fName+": Selected option text: "+sText);

	if (sText == additionalData.showFlagValue){
		let filters=new Object();
		if ("filters" in additionalData) filters=additionalData.filters;
		x.style.display = "block";
		updateListDisplay(additionalData.divQueryName,
			additionalData.queryName,filters,readonlyFlag);
	}
}

//>>have list refresh when data is added (not optimal yet)
//
function isFilterList(v){
   if (typeof(v)!='string') return false;
   if (v.search(';')==-1) return false;
   return true;
}

function updateListDisplay(divName,queryName,filters,readonlyFlag){
	//use Labkey.QueryWebPart to show list

	let debug=true;
	let fName="[updateListDisplay]";

	if (debug) 
		print(fName+": UpdateListDisplay: Query - "+queryName
			+" div - "+divName);

	if (divName=="NONE") return;

	let crfRef=getCRFref();
	let div=config.document.getElementById(divName);

	if (debug)
		print(fName+": generating WebPart: "+queryName);
	
	var qconfig=new Object();
	qconfig.renderTo=divName;
	//point to data container
	qconfig.containerPath=getContainer('data');
	qconfig.schemaName='lists'; 
	qconfig.queryName=queryName;
	qconfig.buttonBarPosition='top';
	qconfig.filters=[];
	for (f in filters){
      let fType=LABKEY.Filter.Types.EQUAL;
      print(fName+' filter ['+f+'] '+filters[f]+'/'+typeof(filters[f])+' ['+fType+']');
     
      if (isFilterList(filters[f])){
         fType=LABKEY.Filter.Types.IN;
      }
		qconfig.filters.push(LABKEY.Filter.create(f, filters[f],fType));
	}
	qconfig.success=updateSuccess;
	qconfig.failure=updateFailure;
	//show only print button
	if (readonlyFlag){
		qconfig.buttonBar=new Object();
		qconfig.buttonBar.items=["print"];
	}

	LABKEY.QueryWebPart(qconfig);
	
}

function updateSuccess(data){
	print("Update success");
}

function updateFailure(data){
	print("Update failed");
}


//TODO: this should trigger a data refresh on section, ie populateData(field)
function toggleVisibility(divName,buttonName){
	let fName='[toggleVisibility/'+divName+']';
	print(fName);

	let x = config.document.getElementById(divName);
	if (x.style.display === "none") {
		//exclude non data sections (like debug)...
		print(fName+': issuing setData(populateSection)');
    		x.style.display = "block";
		config.document.getElementById(buttonName).value="Hide";
		let cb=function(){populateSection(divName);};
		setData(cb);

  	} else {
    		x.style.display = "none";
		config.document.getElementById(buttonName).value="Show";

  	}
}


function generateButton(divName,caption,label,callbackLabel,callback){
	let debug=true;
	if (debug) print("generateButtonX");
	
	let tb=config.document.createElement('table');
	tb.className="t2";
	
	let r1=tb.insertRow();
	th=config.document.createElement('th');
	r1.appendChild(th);
	th.innerHTML=caption;
	//*!*
	let c2=r1.insertCell();
	let i1=config.document.createElement("input");	
	i1.type="button";
	i1.value=label;
	i1.style.fontSize="20px";
	i1.onclick=callback;
	i1.id='button_'+callbackLabel;
	c2.appendChild(i1);	

	let c1=r1.insertCell();
	c1.setAttribute("colspan","1");
	//this is only for saveReview?
	c1.id=divName+'_reportField';
	//c1.id=config.submitReportId;

	let el=config.document.getElementById(divName);
	if (debug) print("generateButton: element["+divName+"]: "+el);
	
	
	el.appendChild(tb);
	
	
}

function generateSubQuery(input, setup, readonlyFlag){
	let fName="[generateSubQuery]";
	if (setup.isReview) return;

	if (!(setup.queryName in config.formConfig.additionalData)){
		print(fName+': no additionalData entry (probably a subquery)');
		return;
	}

	let additionalData=config.formConfig.additionalData[setup.queryName];
	if (!("showFlag" in additionalData))
		return;

	print(fName);
		
	let expId=setup.getInputId(additionalData.showFlag);
	if (expId!=input.id) {
		print(fName+": ignoring field "+input.id+"/"+expId);
		return;
	}

	print(fName+": Setting onChange to "+input.id);
	if (!readonlyFlag)
		input.onchange=function(){setListVisibility(input,setup,readonlyFlag)};
}


//>>populate fields
//
//
//split to field generation and field population
//
function addFieldRow(tb,field,setup,additionalData){

	let fName="[addFieldRow/"+setup.queryName+':'+field.name+']';

	let vName=field.name;
	let vType=field.type;
	let isLookup=("lookup" in field);
	print(fName+": ["+vName+"/"+vType+'/'+isLookup+"]");

	let row=tb.insertRow();
	let cell=config.document.createElement('th');
   cell.style.width='300px';
	row.appendChild(cell);
	
	let text = config.document.createTextNode(field.shortCaption);
	cell.appendChild(text);
		
	
	let input=null;
   let colSpan="3";
	let cell1=row.insertCell();
	cell1.colSpan=colSpan;
	let readonlyFlag=setup.readonlyFlag(vName);


	//set the html input object
	while (1){

		if (readonlyFlag){
			input=config.document.createElement('label');
			input.innerText='Loading';
			break;
		}
	

		//lookup
		if (isLookup){
			input = config.document.createElement("select");
			break;
		}

		//date
		if (vType=="date"){ 
			input = config.document.createElement("input");
			input.type="date";
			break;
		}

		//string
		if (vType=="string"){
			//we have to make sure UNDEF is carried to below
			//since we are adapting file to either show
			//current file or allow user to select a file
			//
			//TODO change this so one can always select file
			//but also show the selected file

			if(vName.search("reviewComment")>-1){
				input = config.document.createElement("textarea");
				input.cols="65";
				input.rows="5";
				break;
			}

			input=config.document.createElement('input');
			input.type="text";
			
			if (vName.search('_file_')<0) break;
			cell1.setAttribute('colspan',"1");
			let cell2=row.insertCell();
			cell2.setAttribute('colspan',"2");
			let input1=config.document.createElement('input');
			input1.type="file";
			input1.id=setup.getInputId(vName)+'_file_';
			cell2.appendChild(input1);
			break;
				
		}


		if (vType=="float"){
			input = config.document.createElement("input");
			input.type="text";
			break;
		}	
		
		
		if (vType=="boolean"){
			input = config.document.createElement("input");
			input.type="checkbox";
			print("Creating checkbox");
			break;
		}
		break;
	}
	
	input.id=setup.getInputId(vName);
	cell1.appendChild(input);
	print(fName+': adding element '+input.id);
	print(fName+': listing element '+config.document.getElementById(input.id));
	

	//connect associated list
	generateSubQuery(input,setup,readonlyFlag);	

	if (readonlyFlag) {
		print(fName+': exiting(readonlyFlag)');
		return;
	}
	
	if (!isLookup) 	{
		print(fName+': exiting (not lookup)');
		return;
	}

	let lookup=field["lookup"];

	//get all values from config.formConfig.lookup[X]
	let lObject=config.formConfig.lookup[lookup.queryName];
	
	//debug
	print(fName+": query: "+lookup.queryName);
	print(fName+": ElementId: "+input.id);
	print(fName+": No of options: " +lObject.LUT.length);
	print(fName+": Element: "+input);

	//set the lut value (input is text label) for readonly

   	//clear existing fields from input	
	for(let i = input.options.length; i >= 0; i--) {
		input.remove(i);
   	}
	
	//create option -1
	let opt = config.document.createElement("option");
	opt.text = "<Select>";
	opt.value = -1;
	input.options[0] = opt;
	print(fName+": Adding <Select>");
	

	//add other, label them with LUT
	for (let v in lObject.LUT) {
		print(fName+': populating '+v+': '+lObject.LUT[v]);

		let opt = config.document.createElement("option");
		opt.text = lObject.LUT[v];
		opt.value = v;
		input.options[input.options.length] = opt;
		
	}
	input.selectedIndex=0;	

}

function parseVariables(pars){
   let pA=pars.split(";");
   let q=new Object();
   for (let i=0;i<pA.length;i++){
      let vA=pA[i].split('=');
      q[vA[0]]=vA[vA.length-1];
   }
   return q;
}

function printVariables(q){
	let fName="[printVariables]";
	for (let x in q){
		print(fName+" ["+x+"] "+q[x]);
	}
}

function hasVariable(q,varName){
   if (q && varName in q)
      return true;

   return false;
}

function addSpecialFieldRow(tb,specFieldSetup,setup){
   //tb is the table, specFieldSetup is a row from the table where special fields are being setup
   //the first column is fieldUID, which is a colon joined amalgation of queryName:fieldName
   let fieldUID=specFieldSetup["fieldUID"];
   let x=fieldUID.split(':');
   let fieldName=x[1];
   let fName="[addSpecialFieldRow/"+fieldUID+"]";
   let q=parseVariables(specFieldSetup['actionParameters']);
   print(fName);
   if (specFieldSetup['actionType']=='textArea'){
      let row=tb.insertRow();
      let cell1=row.insertCell();
      cell1.colSpan="4";
      cell1.style.textAlign="justify";
      cell1.style.padding="10px";
      cell1.style.backgroundColor="#e0e0e0";
      cell1.innerText=q['description'];
      return;
   }
   if (specFieldSetup['actionType']=='generationObject'){
      //only in EDIT mode!!
      let ro=setup.readonlyFlag(fieldName);
      if (ro) return;
      config.print=print;
      let gc=getGenerationObject(config,q,setup.getInputId(fieldName));
      if ('mailRecipient' in q){
         gc.callback=function(data){sendEmail(data,q['mailRecipient'],doNothing,q['subject']);};
      }
      if ("addData" in q){
         vars=q["addData"].split(',');
         gc.addData=new Array();
         for (let v in vars){
            let s=vars[v]
            //variable name can be written as A/B where A is the name in addData and B is the variable name in crfEntry
            //useful for mocking up crfId from daughter crf-s such as registration
            let sArray=s.split('/');
            let sTarget=sArray[0];
            let sSource=sArray[sArray.length-1];
            gc.addData[sTarget]=config.formConfig.crfEntry[sSource];
            print(fName+" addData ["+sTarget+"]: "+gc.addData[sTarget]);
         }
      }
      let row=tb.insertRow();
      let cell=config.document.createElement('th');
      row.appendChild(cell);
      let text = config.document.createTextNode("Automatic ID generator");
      cell.appendChild(text);
      let cell1=row.insertCell();
      cell1.colSpan="3";
      let b=config.document.createElement("input");
      b.type="button";
      b.id="generateIdButton";
      b.onclick=function(){gc.execute();};
      b.value="Generate ID";
      cell1.appendChild(b);

   }
}

		
function populateFieldRow(entry,field,setup){
	populateField(entry,field,setup);
	populateSubQuery(entry,field,setup);
}

function populateSubQuery(entry,field,setup){
	let fName='[populateSubQuery/'+setup.queryName+':'+field.name+']';
	if (setup.isReview) return;
	
	if (!(setup.queryName in config.formConfig.additionalData)){
		let msg=fName+': no additionalData entry for '+setup.queryName;
		msg+=' (probably a subquery)';
		print(msg);
		return;
	}
	//find if field is connected to a sub array
	//find queryName
	//
	let additionalData=config.formConfig.additionalData[setup.queryName];	
	print(fName);
	//let flag=additionalData.showFlag;
	
	if (!("showFlag" in additionalData)) return;
	let eId=setup.getInputId(additionalData.showFlag);
	let id=setup.getInputId(field.name);
	
	if (eId!=id) {
		print(fName+": ignoring field "+id+"/"+eId);
		return;
	}
	
	print(fName+': id '+id);
	//hard to estimate readonlyFlag
	//
	let input=config.document.getElementById(id);
	let eType=input.nodeName.toLowerCase();
	let readonlyFlag=eType!="select";
	setListVisibility(input,setup,readonlyFlag);

}

function populateField(entry,field,setup){

	let vName=field.name;
	let fName='[populateFieldName/'+vName+']';

	let varValue="UNDEF";

	//if (vName in setup.filters) varValue=setup.filters[vName];
	if (vName in entry) varValue=entry[vName];
	//if part of the filter, set it to value
	if (vName in setup.filters) varValue=setup.filters[vName];
	
	let isLookup=("lookup" in field);
	
	print(fName+' v='+varValue+'/'+isLookup+' ['+
		setup.getInputId(field.name)+']');
	
	let vType=field.type;
	let id=setup.getInputId(vName);
	let input=config.document.getElementById(id);

		
	//date
	if (vType=="date"){
		if (varValue=="UNDEF") varValue=new Date();
		else varValue=new Date(varValue);
	}
	
	//lookup for readonly
	if (isLookup && varValue!="UNDEF"){
		let lookup=field["lookup"];
		//get all values from config.formConfig.lookup[X]
		let lObject=config.formConfig.lookup[lookup.queryName];
		varValue=lObject.LUT[varValue];
	}

	print('Element: '+input);
	//figure out the element type
	let eType=input.nodeName.toLowerCase();
	print('Element type: '+eType);

	//change varValue for printing
	if (varValue=="UNDEF") varValue="";
	//HTMLTextArea, createElement(textArea)
	if (eType==="textarea"){
		input.value=varValue;
		return;
	}
	//Text, createTextNode
	if (eType==="#text"){
		input.nodeValue=varValue;
		return;
	}
	//HTMLLabelElement, createElement('label')
	if (eType==="label"){
		input.innerText=varValue;
		return;
	}

	//HTMLSelectElement, createElement('select')
	if (eType==="select"){
		input.selectedIndex=0;
		for (let i=0;i<input.options.length;i++){
			let v=input.options[i].text;
			if (v!=varValue) continue;
			input.selectedIndex=i;
			break;
		}
		return;
	}

	if (eType!="input"){
		print('Unknown type: '+eType+' encountered, igonring');
		return;
	}
	
	//HTMLInputElement
	let type=input.type;

	if (type=="date"){
		input.valueAsDate=varValue;
		return;
	}
	//string,float
	if (type=="text"){
		input.value=varValue;
		return;
	}
	//boolean
	if (type=="checkbox"){
		input.checked=varValue;
		return;
	}
	print('Unknown input type: '+type+'. Ignoring.');
}


function populateTable(listName,writeMode){
//function populateTable(formSetupEntry){
	//let listName=config.formConfig.queryMap[formSetupEntry['queryName']];
	//let accessMode=config.formConfig.operator+'Mode';
	//let writeMode=formSetupEntry[accessMode]=='EDIT';

	let fName='[populateTable/'+listName+']';

	let setup=getSetup(listName,writeMode);
	let entry=new Object();
	
	//data snapshot
	let fQuery=getQuerySnapshot(listName);
   let queryLayout=getQueryLayout(listName);

	//here I assume that listName was parsed during setDataLayout and setData 
	//so that rows was set (even if they are empty)
	print(fName+"]: nrows "+fQuery.rows.length);
	
	if (fQuery.rows.length>0)
		entry=fQuery.rows[0];
	
	let fields=queryLayout.fields;
		
	for (f in fields){	
		let field=fields[f];
		//each field is a new row
		print(fName+": Adding field: "+f+'/'+field.name+' hidden: '+field.hidden+' type:'+field.type);
		if (field.hidden) continue;
		if (field.name=="crfRef") continue;
		populateFieldRow(entry,field,setup);
		
	}

}

function generateTable(listName,divName,additionalData,setup){
	let debug=true;
	let fName="[generateTable/"+listName+"]";	
	if (debug) print(fName);

	//is listName and setup.queryName a duplicate of the same value
	print(fName+': setup.queryName '+setup.queryName);	
	//assume data is set in config.formConfig.dataQueries[data.queryName].rows;
   let populateData=true;
   if ("subTable" in setup){
      print(fName+" is subTable");
      populateData=false;
   }

	let entry=new Object();



	//data snapshot
	let fQuerySnapshot=getQuerySnapshot(listName);
   let queryLayout=getQueryLayout(listName);
	//here I assume that listName was parsed during setDataLayout and setData 
	//so that rows was set (even if they are empty)
	print(fName+": Nrows "+fQuerySnapshot.rows.length);
	
	if (fQuerySnapshot.rows.length>0)
		entry=fQuerySnapshot.rows[0];

   
   if ("reviewTable" in setup){
      entry['reviewComment']='';
      delete entry["ModifiedBy"];
   }
	
	let tb=config.document.createElement('table');
	tb.className="t2";
	config.document.getElementById(divName).appendChild(tb);

	//this are the fields (probably constant)
	let fields=queryLayout.fields;
		
	for (f in fields){
		let field=fields[f];
      let fieldUID=listName+":"+field.name;
		//each field is a new row
		print(fName+": Adding field: "+f+'/'+field.name+' ('+fieldUID+').');
      //unique name
		if (field.hidden) continue;
		if (field.name=="crfRef") continue;
		addFieldRow(tb,field,setup,additionalData);
		if (populateData) populateFieldRow(entry,field,setup);
      if (fieldUID in config.formConfig["specialFields"]){
         let specFieldSetup=config.formConfig["specialFields"][fieldUID];
         addSpecialFieldRow(tb,specFieldSetup,setup);
      }

		
	}
   //finish of if apply button is not required
	if (!("addApply" in setup)) {
		print(fName+"populateTable: done");
		return;
	}
	
	let row=tb.insertRow();

	let th=config.document.createElement('th');
	row.appendChild(th);
	th.innerHTML=setup.addApply; 
	let cell=row.insertCell();
	//cell.setAttribute("colspan","2");
	let input=config.document.createElement("input");	
	input.type="button";
	input.value=setup.addApply;
	cell.appendChild(input);	
	let cell1=row.insertCell();
	cell1.setAttribute("colspan","2");
	cell1.id=setup.getInputId("rerviewLastSave");
	cell1.innerHTML="No recent update";
	//saveReview is a generic name for saving content of the html page to a list entry
	input.onclick=function(){saveReview(listName,cell1.id,setup)};
}	
	
function setEntryFromElement(entry,elementId, field){
   //set value to entry from element using representation (field) from labkey
   //
   //
   
   let fName='setEntryFromElement';

   let el=config.document.getElementById(elementId);
				
   if (!el) {
      print(fName+" element: "+elementId+" not found");
      return;
   }
   print(fName+" element: "+elementId);
      

   let vName=field.name;
   let vType=field.type;

   let eType=el.nodeName.toLowerCase();

   if (eType==="select"){
      entry[vName]=el.options[el.selectedIndex].value;
      return;
   }

   if (eType==="td"){
      entry[vName]=el.innerText;
      return;
   }
   
   if (vType=="date"){
      let date=el.valueAsDate;
      if (date==="null") return;

      date.setUTCHours(12);
      entry[vName]=date.toString();
      print(fName+" setting date to "+entry[vName]);
      return;
   }

   if (vType=="string"){
      entry[vName]=el.value;
      
      if (vName.search('_file_')<0) 
         return;
      
      //upload file
      let id1=elementId+'_file_';
      let input1=config.document.getElementById(id1);
      print(fName+' attachment field: '+input1.value);
      //entry[vName]=el.files[0].stream();
      let ctx=new Object();
      ctx['dirName']='consent';
      ctx['ID']=entry['crfRef'];
      //should point to data container
      ctx['project']=getContainer('data');
      //need ID->crf!
      //assume crfRef will get set before this
      //element is encountered
      uploadFile(input1,ctx);
      let fv=el.value;
      let suf=fv.split('.').pop();
      entry[vName]=entry['crfRef']+'.'+suf;
      return;
      
   }	
   if (vType=="float" || vType=="int"){
      entry[vName]=el.value;
      
      if (vName=="queryName") {
         print(fName+' parsing queryName: '+el.innerText);
         entry[vName]=config.formConfig.fields[el.innerText].queryId;
         //use queryMap lookup
      }
      return;
   }	
   if (vType=="boolean"){
      entry[vName]=el.checked;
      return;
	}
   return;
}

function saveReview(queryName,elementId,setup){
	//loads any queryName

	let debug=true;
   let fName='[saveReview/'+queryName+']';
	print(fName+" elementId "+elementId);

   let mode='update';

   let unique=("unique" in setup);
	
	if (!unique) mode='insert';

   //data snapshot
	let fQuerySnapshot=getQuerySnapshot(queryName);
	let queryLayout=getQueryLayout(queryName);

	if (fQuerySnapshot.rows.length==0) useInsert=true;
	let entry=new Object();
	
	if (unique){
		entry=fQuerySnapshot.rows[0];
 	}
	entry.crfRef=getCRFrefData();

	if (debug) print("Set crfRef="+entry.crfRef);

	//if ("queryName" in setup.filters) {
	//	entry.queryName=setup.filters["queryName"];
	//	if (debug) print("Setting queryName: "+entry.queryName);
	//}

	let fields=queryLayout.fields;
	for (f in fields){

		let field=fields[f];
		if (debug) print("saveReview field: "+field.name);
		if (field.hidden) continue;
		
		let vName=field.name;
		let vType=field.type;

		if (debug) print("vType: "+vType);
		
		if (vName=="crfRef") continue;
		//need to save queryName for reviewComments
		
		let eId=setup.getInputId(vName);
      //copy values from form to entry
      setEntryFromElement(entry,eId,field);
      if (!unique){
         let foo=Object();
         //reset the field value
         populateField(foo,field,setup);
      }
	}
	let action=function(data){updateLastSavedFlag(data,setup,elementId)};

	print(fName+" mode "+mode);
   cvModifyRows(mode,'lists',queryName,[entry],action,getContainer('data'));

}

function updateLastSavedFlag(data,setup,elementId){
	let debug=true;
	if (debug) print("Update last saved flag to "+elementId);
	let el=config.document.getElementById(elementId);
	let dt=new Date();
	el.innerHTML="Last saved "+dt.toString();
	if (data.queryName=="reviewComments"){
		updateListDisplay(setup.divReviewListId,"reviewComments",setup.filters,true);
	}	
	//refresh stored data!
	let writeMode=!setup.readonlyFlag();
	if ("unique" in setup)
		setData(function (){populateTable(data.queryName,writeMode);});
	if ("masterQuery" in setup){
		let ad=config.formConfig.additionalData[setup.masterQuery];
		print('Updating list display: '+setup.queryName+'/'+ad.queryName);
		updateListDisplay(ad.divQueryName,ad.queryName,ad.filters,false);
	}
}

//******************************************upload to database *********************

function onDatabaseUpload(){
	let fName='[onDatabaseUpload]';
	print(fName);
	config.upload=new Object();
   let fc=new Object();
   let pM=getParticipantManagerObject(config);
   fc.participantId=pM.getParticipantIdFromCrfEntry();
	print(fName+' id '+fc.participantId);
   afterParticipantId(fc);
}


function afterParticipantId(fc){
	print("Setting participantId to "+fc.participantId);
	config.upload.participantId=fc.participantId;
	//another select rows to update all queries from setup
	//just use registration for test
	let formSetupRows=config.formConfig.formSetupRows;
	config.upload.queries=new Array();
	print("Form rows: "+formSetupRows.length);
	for (let i=0;i<formSetupRows.length;i++){
		let entry=formSetupRows[i];
		//skip reviews
		if (entry.showFlag=="REVIEW") continue;
		//use lookup table to convert from id to name
		let queryName=config.formConfig.queryMap[entry.queryName];
		config.upload.queries.push({queryName:queryName,queryStatus:"QUEUED"});
		print('form ['+i+']='+queryName+' '+entry.showFlag+'/'+entry.showQuery);
		if (entry.showQuery=="NONE")
			continue;
		config.upload.queries.push({queryName:entry.showQuery,queryStatus:"QUEUED"});
	}
	//add reviews
	config.upload.queries.push({queryName:"reviewComments",queryStatus:"QUEUED"});
	config.upload.queryId=0;
	copyToDataset();

}



function copyToDataset(){
	let fName='[copyToDataset]: ';
	print(fName+'['+config.upload.queryId+'/'+config.upload.queries.length+']');
	//watch dog + scheduler
	//

	//watchdog part
	if (config.upload.queryId==config.upload.queries.length) {
		print(fName+'completing');
		let targetStatus=config.formConfig.targetStatus['onDatabaseUpload'];
		let targetRecipient=config.formConfig.targetRecipient['onDatabaseUpload'];
		let action=new Object();
		action.name='onDatabaseUpload';
		action.cb=function(data){sendEmail(data,targetRecipient,redirect,'Form uploaded');}
		updateFlag(targetStatus,action);//Approved
		return;
	}

	//scheduler
	let queryName=config.upload.queries[config.upload.queryId].queryName;
	print("copyToDataset["+config.upload.queryId+"/"+
			config.upload.queries.length+"]: "+queryName);

	let filters=[LABKEY.Filter.create('crfRef',getCRFref())];
   cvSelectRows('lists',queryName,filters,afterListData,getContainer('data'));
}

function afterListData(data){
	let fName='[afterListData]: ';

	let queryName=config.upload.queries[config.upload.queryId].queryName;
	print(fName+" ["+queryName+"/list]: "+data.rows.length+" entries");
	config.upload.queries[config.upload.queryId].listData=data;
	let id=config.upload.participantId;

	let filters=[LABKEY.Filter.create('crfRef',getCRFref()),LABKEY.Filter.create('ParticipantId',id)];
   cvSelectRows('study',queryName,filters,afterStudyData,getContainer('data'));
}


function afterStudyData(data){

	let fName='[afterStudyData]: ';
	let queryObj=config.upload.queries[config.upload.queryId];
	queryObj.studyData=data;

	let msg=fName+"["+queryObj.queryName+"/study]: "+data.rows.length+" entries";
	print(msg);
	

	let listRows=queryObj.listData.rows;
	//skip uploading an empty set
	if (listRows.length==0){
		printErr("List "+queryObj.queryName+" empty.");
		queryObj.queryStatus="DONE";
		config.upload.queryId+=1;
		//back to watchdog
		copyToDataset();
		return;
	}
	

	let studyRows=queryObj.studyData.rows;

	for (let i=0;i<studyRows.length;i++){
		let entry=studyRows[i];
		//
		if (! (i<listRows.length) ) continue;
		let entryList=listRows[i];
		//keeps study only variables (ParticipantId, SequenceNum)
		for (let f in entryList) {
			entry[f]=entryList[f];
			print(fName+"Copying ["+f+"]: "+entry[f]+"/"+entryList[f]);
		}
	}
	print(fName+' copying completed');

	if (studyRows.length>0) {
      cvModifyRows('update','study',queryObj.queryName,studyRows,afterStudyUpload,getContainer('data'));
		print(fName+'updateRows sent');

	}
	else{
		let data=new Object();
		data.rows=new Array();
		afterStudyUpload(data);
	}
}

function afterStudyUpload(data){
	let fName='[afterStudyUpload] ';
	print(fName);
	//let participantField=config.participantField;
	let participantField=config.formConfig.studyData["SubjectColumnName"];
	print(fName+' participantField: '+participantField);

	let queryObj=config.upload.queries[config.upload.queryId];
	let queryName=queryObj.queryName;
	printErr("Updated "+data.rows.length+" rows to "+queryName);
	
	let studyRows=queryObj.studyData.rows;
	let listRows=queryObj.listData.rows;
	
	let rows=new Array();
	//also updating existing rows, if they exist
	for (let i=studyRows.length;i<listRows.length;i++){
		let entry=listRows[i];
		//make sure you have the participantField right
		//
		entry[participantField]=config.upload.participantId;
		entry.crfRef=getCRFref();
		entry.SequenceNum=getCRFref();
		entry.SequenceNum=entry.SequenceNum % 1000000000;
		
		if (listRows.length>1){
			entry.SequenceNum+=i/100;
		}
		print( "Adding sequence number "+entry.SequenceNum);
		rows.push(entry);
	}
	if (rows.length>0){
      cvInsertRows('study',queryName,rows,afterListUpload,getContainer('data'));
	}
	else{
		let data=new Object();
		data.rows=rows;
		afterListUpload(data);
	}
			
}

function afterListUpload(data){
	
	let queryObj=config.upload.queries[config.upload.queryId];
	let queryName=queryObj.queryName;
	printErr("Inserted "+data.rows.length+" rows to "+queryName);
	queryObj.queryStatus="DONE";
	config.upload.queryId+=1;
	copyToDataset();

}




//*************************update for further review *************************

function onUpdateForReview(){
	let targetStatus=config.formConfig.targetStatus['onUpdateForReview'];
	let targetRecipient=config.formConfig.targetRecipient['onUpdateForReview'];
	let action=new Object();
	action.name='onUpdateForReview';
	action.cb=function(data){sendEmail(data,targetRecipient,redirect,'Form updated for review');};

	updateFlag(targetStatus,action);
}

function updateFlag(flag,action){
	let fName='[updateFlag 1]';
	let debug=true;

	let entry=config.formConfig.crfEntry;
	entry.FormStatus=flag;
	let uId=config.formConfig.currentUser.UserId;
	entry[config.formConfig.operator]=uId;

	print(fName+': Form: '+entry.Form);
	print(fName+": set form status to "+entry.FormStatus);
	
	let cb=function(data){completeWithFlag(data,action);};
   cvModifyRows('update','lists','crfEntry',[entry],cb,getContainer('data'));
		
}



function completeWithFlag(data,action){
	let fName='[completeWithFlag]';
	print(fName+': nrows '+data.rows.length);

	let fentry=data.rows[0];
	print(fName+': form status '+fentry.FormStatus);
	print(fName+': form '+fentry.Form);

	let crfStatus=createCrfStatus(fentry);
	crfStatus.operator=config.formConfig.operator;
	crfStatus.action=action.name;

   cvInsertRows('lists','crfStatus',[crfStatus],action.cb,getContainer('data'));

}

//************************************************ submit *******************************************

function onSubmit(){
	//update list storage and change status

	hideErr();
	clearErr();
	printErr("onSubmit");

	setData(verifyData);
	


}

function verifyData(){
	let fName='[verifyData]';
   let qList=getQueryList();
	for (q in qList){
		let qData=getQuerySnapshot(q);
		if (q=="reviewComments") continue;
		//copy snapshot to history
      if (qData.rows.length==0){
         print(fName+' no rows for '+q);
      }
      else
		   cvInsertRows('lists',q+'History',qData.rows,doNothing,getContainer('data'));
		//if it doesn't have additionalData, it is a sub query
		if (!(q in config.formConfig.additionalData)){
			continue;
		}
		if (qData.rows.length<1){
			printErr('Missing entry for query '+q);
			return false;
		}
	}
	//this is necessary only for Generated to Generation completed step
	let actionSettings=config.formConfig.actionSettings['onSubmit'];	
	if (hasVariable(actionSettings,"updateRegistration")){
		updateRegistration();
	}
	let targetStatus=config.formConfig.targetStatus['onSubmit'];	
	let targetRecipient=config.formConfig.targetRecipient['onSubmit'];
	print(fName+' targetStatus: '+targetStatus);
	
	let finalStep=redirect;
	if (hasVariable(actionSettings,"finalStep")){
		//set to doNothing to remain on submit window
		if (actionSettings.finalStep=="doNothing"){
			finalStep=doNothing;
		}
	}
	
	let action=new Object();
	action.name='onSubmit';
	action.cb=function(data){sendEmail(data,targetRecipient,finalStep,'Form sumbitted');};
	updateFlag(targetStatus,action);
}

function getEmail(recipientCode){

	print('getEmail w/'+recipientCode);
	let recipients=new Array();
	let typeTo=LABKEY.Message.recipientType.to;
	let create=LABKEY.Message.createRecipient;
	let currentUser=config.formConfig.currentUser;
	let formCreator=config.formConfig.formCreator;
	let currentSite=config.formConfig.currentSite;
	let userRows=config.formConfig.userRows;
	let parentUser=undefined;
	if ("parentCrfData" in config.formConfig){
		let parentCrf=config.formConfig.parentCrfData;
		parentUser=getUser(parentCrf.rows[0].UserId,'parentUser');
	}
	

	let recipientCategories=recipientCode.split(',');
	for (let i=0;i<recipientCategories.length;i++){

		let recipient=recipientCategories[i];
		print('Checking '+recipient);
		if (recipient=='crfEditor'){
			print('Adding :'+formCreator.Email);
			recipients.push(create(typeTo,formCreator.Email));
			if (parentUser==undefined) continue;
			print('Adding :'+parentUser.Email);
			recipients.push(create(typeTo,parentUser.Email));
			continue;
		}
		//Monitor or Sponsor
		let fList=recipient+'s';
		let fRows=config.formConfig[fList];
		for (let i=0;i<fRows.length;i++){
			print('Checking '+fRows[i].User+'/'+fRows[i].Site);
			if (fRows[i].Site!=currentSite.siteNumber) continue;
			for (let j=0;j<userRows.length;j++){
				if (userRows[j].UserId!=fRows[i].User) continue;
				print('Adding :'+userRows[j].Email);
				recipients.push(create(typeTo,userRows[j].Email));
				break;
			}
		}
	}

	return recipients;
}

function sendEmail(data,recipient='crfEditor',cb=redirect,subj='Form submitted'){

	print('sendEmail; recipient: '+recipient);

	let st=config.formConfig.settings;
	let cvar='sendEmail';
	if (cvar in st){
		print(cvar+' set to '+st[cvar]);
		if (st[cvar]=='FALSE'){
			print('Skipping sending emails');
			cb();
			return;
		}
	}
	if (recipient==null){
      print('Skipping sending emails w/ no recipients');
      cb();
      return;
   }


	print('send email '+data.rows.length);
	let crf=data.rows[0]['entryId'];
	let formId=data.rows[0]['Form'];
	let link=LABKEY.ActionURL.getBaseURL();
	link+=LABKEY.ActionURL.getContainer();
	link+='/crf_tecant-visit.view?';
	link+='entryId='+crf;
	link+='&formId='+formId;
	link+='&role='+recipient;

	//debug
	let recipients=getEmail(recipient);
	//from crfManagers list
	
	let typeHtml=LABKEY.Message.msgType.html;
	let typePlain=LABKEY.Message.msgType.plain;
	let msg1=LABKEY.Message.createMsgContent(typePlain,link);

	//let cb=doNothing;
	//let cb=redirect;
	LABKEY.Message.sendMessage({
		msgFrom:'labkey@fmf.uni-lj.si',
		msgSubject:subj,
		msgRecipients:recipients,
		msgContent:[msg1],
		success: cb
	});

}

function hideErr(){
	let el=config.document.getElementById("errorDiv");
	el.style.display="none";
}

function clearErr(){
	let el=config.document.getElementById("errorTxt");
	el.value="";
}

function showErr(){
	let el=config.document.getElementById("errorDiv");
	el.style.display="block";
}

function printErr(msg){
	showErr();
	el=config.document.getElementById("errorTxt");
	el.style.color="red";
	el.value+="\n"+msg;
}


//**************************************************
//
function onRemoveCRF(){
   let fName='[onRemoveCRF]';
	config.inputListsIterator=0;
   print(fName+' starting loop');

   //let rd=function(data){redirect();};
   //let cb=function(){cvInsertRows('lists','crfStatus',[crfStatus],rd,getContainer('data'));};
   let cb=function(){removeCrfEntries(redirect);};
	removeCRFLoop(cb);
}

function removeCRFLoop(cb){
   let fName='[removeCRFLoop()]';
	let debug=true;

	let i=config.inputListsIterator;
	let iMax=config.formConfig.inputLists.rows.length;
	//in fact, we are adding two additional passages of the loop, one for 
	//crfEntry, the second for the same query, but using parentCrf as the 
	//selection variable
	//let iTotal=iMax+1;
	//let iTotal=iMax+1;
   //
   let actionSettings=config.formConfig.actionSettings['onRemoveCRF'];
   let queryNameDeleteWithParentCrf='NONE';
   if (hasVariable(actionSettings,'removeWithParentCrf')){
      queryNameDeleteWithParentCrf=actionSettings['removeWithParentCrf'];
   }
	
	print(fName+" ["+i+"/"+iMax+"]");

	if (!(i<iMax)){
		if (0) return;
      cb();
	}
	//in all but crfEntry, variable is called crfRef
	let queryName=config.formConfig.inputLists.rows[i].queryName;
   let idVar="crfRef";
   let idValue=config.formConfig.crfEntry['entryId'];

	//delete also crfEntries where parentCrf is set to crf that we are deleting

   if (queryNameDeleteWithParentCrf==queryName){
      idValue=config.formConfig.crfEntry['parentCrf'];
   }
	
	print(fName+" ["+i+"/"+iMax+"] "+queryName+":"+idVar+'/'+idValue);

   let filters=[LABKEY.Filter.create(idVar,idValue)];
   let action=function(data){removeListCRF(data,cb);};
   let failure=function(errorInfo){skipListCRF(errorInfo,cb);};
   cvSelectRows('lists',queryName,filters,action,getContainer('data'),failure);
	//selectRows.failure=skipListCRF;
}

function removeListCRF(data,cb){
   let fName="[removeListCRF]";
	print(fName+" "+data.queryName+": "+data.rows.length);

	config.inputListsIterator+=1;
	
	if (data.rows.length==0){
		removeCRFLoop(cb);
		return;
	}

   let action=function(data){removeCRFLoop(cb)};
   cvDeleteRows(data.schemaName,data.queryName,data.rows,action,getContainer('data'));
	
}

function skipListCRF(errorInfo,cb){
   let fName='[skipListCRF]';
	print(fName+" error in removeCRF: "+errorInfo.exception);

	config.inputListsIterator+=1;
	
	removeCRFLoop(cb);
	
}

function removeCrfEntries(cb){
	let queryName="crfEntry";
	let idVar="entryId";
	let crfRef=getCRFref();

   let filters=[LABKEY.Filter.create('entryId',crfRef)];
   let action=function(data){deleteAndUpdateCrfStatus(data,null);}
   cvSelectRows('lists',queryName,filters,action,getContainer('CRF'));
   let action1=function(data){deleteAndUpdateCrfStatus(data,cb);}
   let filters1=[LABKEY.Filter.create('parentCrf',crfRef)];
   cvSelectRows('lists',queryName,filters1,action1,getContainer('CRF'));
}

function deleteAndUpdateCrfStatus(data,cb){
   let fName='[deleteAndUpdateCrfStatus]';
   let rows=data.rows;
   let stack=new Array();
   stack.push(cb);
   for (let i=0;i<rows.length;i++){
      //generate crfStatus entry out of crfEntry
      let crfStatus=createCrfStatus(rows[i]);
      crfStatus.action='onRemoveCRF';
      crfStatus.FormStatus=config.formConfig.targetStatus[crfStatus.action];
      print(fName+' status '+crfStatus.FormStatus);
      crfStatus.operator=config.formConfig.operator;
      let k=stack.length-1;
      stack.push(function(){cvInsertRows('lists','crfStatus',[crfStatus],stack[k],getContainer('CRF'));});
      let k1=k+1;
      stack.push(function(){cvDeleteRows('lists','crfEntry',[rows[i]],stack[k1],getContainer('CRF'));});
   }
   //execute the whole stack
   let m=stack.length-1;
   stack[m]();
}


function redirect(){

	let debug=false;
	let formUrl="begin";
	let params=new Object();
	params.name=formUrl;
	params.pageId="CRF";

	//points to crf container
	let containerPath=getContainer('CRF');
        
	// This changes the page after building the URL. 
	//Note that the wiki page destination name is set in params.
        
	var homeURL = LABKEY.ActionURL.buildURL(
			"project", formUrl , containerPath, params);
        print("Redirecting to "+homeURL);
	if (debug) return;	 
	window.location = homeURL;

	

}

//master section, entry point from html files

function generateMasterForm(){

	generateDebugSection();
	//read enviroment from lists disperesed on labkey
	setFormConfig();
}


//helper function to set basic parameters on web page 
//(fields defined in html file)
function populateBasicData(){

   let staticData=new Object();
   let titles=new Object();
   staticData['version']=config.formConfig.softwareVersion;	
   titles['version']='Software version';
   let varRows=config.formConfig['crfStaticVariables'].rows;
   for (let i=0;i<varRows.length;i++){
      let vName=varRows[i].staticVariable;
      let val=config.formConfig.crfEntry[vName];
      if (val==undefined) continue;
      staticData[vName]=val;
      titles[vName]=varRows[i].Title;
   }
	staticData['investigatorName']=config.formConfig.user['DisplayName'];
   titles['investigatorName']='Investigator';
   staticData['email']=config.formConfig.user['Email'];
   titles['email']='Email';
   staticData['siteName']=config.formConfig.currentSite['siteName'];
   titles['siteName']='Site';
   staticData['sitePhone']=config.formConfig.currentSite['sitePhone'];
   titles['sitePhone']='Telephone(site)';

   for (f in staticData){
      addStaticData(f,titles[f],staticData[f]);
   }
}

function addStaticData(f,title,value){
   let el=config.document.getElementById(f);

   //populate only
   if (el!=undefined){
      el.innerText=value;
      return;
   }
   
   //add row to table if element cannot be found
   let table=config.document.getElementById('staticTable');
   let row=table.insertRow();
   let cell=row.insertCell();
   cell.innerText=title;
   let cell1=row.insertCell();
   cell1.id=f;
   cell1.style.fontWeight='bold';

   //populate
   cell1.innerText=value;
}

//come here after the layout is read from labkey page
//
function generateErrorMsg(msg){
	let txt=config.document.createElement('p');
	txt.innerText=msg;
	config.document.getElementById(config.masterForm).appendChild(txt);
	generateButton("submitDiv",'Exit','Exit','redirect',redirect);
}

function getUser(id,field){
	if (field in config.formConfig) return config.formConfig[field];
	let uRows=config.formConfig.userRows;
	for (let i=0;i<uRows.length;i++){
		let userId=uRows[i].UserId;
		if (userId!=id) continue;
		config.formConfig[field]=uRows[i];
		return config.formConfig[field];
	}
	return null;
}	


function afterConfig(){

	let debug=true;

	if (debug)
		print("afterConfig");	

	populateBasicData();
	
	//check if user has permission on the form
	let currentUser=getUser(LABKEY.Security.currentUser.id,'currentUser');
	let currentSite=config.formConfig.currentSite;
	let formCreator=getUser(config.formConfig.crfEntry.UserId,'formCreator');
	let formCreatorId=formCreator.UserId;



	//let formSite=config.formConfig.crfEntry.Site;
	let fList=config.formConfig.operator+'s';
	let fRows=config.formConfig[fList];
	//let currentSiteId=-1;
	
	//depending on operator mode, we should decide what is right
	let operator=config.formConfig.operator;
	if (operator=='crfEditor'){
		//editor can only edit its own forms
		if (currentUser.UserId!=formCreatorId){
			let msg='User '+currentUser.DisplayName;
			msg+=' has no permission on this form';
			generateErrorMsg(msg);
			return;
		}
	}
	if (operator=='crfMonitor' || operator=='crfSponsor'){
		//monitor can look at forms based on his site
		//find monitor line
		let operatorSites=new Array();
		for (let i=0;i<fRows.length;i++){
			if (fRows[i].User!=currentUser.UserId) continue;
			operatorSites.push(fRows[i].Site);
		}
		print('operator Site: '+operatorSites.length);
		if (operatorSites.length==0){
			let msg='User '+currentUser.DisplayName;
			msg+=' is not a '+operator;
			generateErrorMsg(msg);
			return;
		}

		let selectedSite=-1;
		let siteCandidates="[";
		for (let i=0;i<operatorSites.length;i++){
			if (i>0) siteCandidates+=", ";
			siteCandidates+=operatorSites[i];
			if (operatorSites[i]!=currentSite.siteNumber) continue;
			selectedSite=currentSite.siteNumber;
			break;
		}
		siteCandidates+="]";
		if (selectedSite==-1){
			let msg='User '+currentUser.DisplayName;
			msg+=' is not a '+operator+' for site ';
			msg+=currentSite.siteName+'('+currentSite.siteNumber+')';
			msg+='/'+siteCandidates;
			generateErrorMsg(msg);
			return;
		}
	}


	print('User '+currentUser.DisplayName+'/'+
		config.formConfig.currentSite['siteName']+
		' acting as '+config.formConfig.operator);	


	let rows=config.formConfig.crfButtons.rows;
	config.formConfig.targetStatus=new Array();
	config.formConfig.targetRecipient=new Array();
	config.formConfig.actionSettings=new Array();

	for (let i=0; i<rows.length; i++){
		let action=rows[i].action;//String
		let tstatus=rows[i].targetFormStatus;
		let trecip=rows[i].targetRecipient;
		config.formConfig.targetStatus[action]=tstatus;
		config.formConfig.targetRecipient[action]=trecip;
		//allow for settings to be promoted with each action (and potentially parsed and acted upon)
		config.formConfig.actionSettings[action]=undefined;
		let aSet=rows[i].actionSettings;
		if (aSet){
			config.formConfig.actionSettings[action]=parseVariables(aSet);
			printVariables(config.formConfig.actionSettings[action]);
		}

	}
	let formStatus=config.formConfig.formStatus;

	//let functionArray=new Array();

	print("Generating buttons for formStatus \""+ formStatus+"\"");

   let allButtonRows=config.formConfig.crfButtons.rows;
	let buttonRows=new Array();
   
   //specifying role=X in actionSettings will limit button to that role
   for (let i=0;i<allButtonRows.length;i++){
      let action=allButtonRows[i]['action'];
      //filter on actionSettings
      let as=config.formConfig.actionSettings[action];
      if (hasVariable(as,'role')){
         print('Role['+config.formConfig.operator+'/'+as['role']+'] limited for action '+action);
         //mismatch skips addition of button to buttonRows
         if (config.formConfig.operator!=as['role']) continue;
      }
      buttonRows.push(allButtonRows[i]);
   }



	for (let i=0;i<buttonRows.length;i++){
		let bt=buttonRows[i];
		if (typeof window[bt.action]==="function"){
			generateButton("submitDiv",bt.caption,bt.label,bt.action,
				window[bt.action]);
		}
		else{
			print('No match for function :'+bt.action+
				' obj: '+window[bt.action]);
		}
	}

	print('Here');


	//here we should get data. For now, just initialize objects that will hold data
	setDataLayout(afterDataLayout);//callback is afterDataLayout
}
	
function afterDataLayout(){

	setData(afterData);//callback is afterData
}

function verifyCrfStudyId(pM){

   //is studyId already set for the crf
   let studyId=pM.getParticipantIdFromCrfEntry('STUDY');
   if (!studyId) return;
   pM.mode="STUDY";
   pM.readOnly="TRUE";
}

function verifyRegistration(pM){
   //if registration is in, 
   //then local id should not be changed any longer
   let idFieldName=pM.getCrfEntryFieldName("STUDY");
   //let registrationQuery=config.formConfig.settings['registrationQuery'];
   //if (!registrationQuery) return; //LOCAL is OK
   //let fQuery=config.formConfig.dataQueries[registrationQuery];
   //if (!fQuery) return; //no registration query, then it should be ignored
   let fQuery=config.formConfig.registrationData;
   if (fQuery.rows.length==0) return; //registration is empty

   let studyId=fQuery.rows[0][idFieldName];
   if (!studyId) return; //study id not set
   //set 
   pM.mode="STUDY";
   pM.readOnly="TRUE";
   //set crf (this happens later, but probably before the form will be corrected)
   pM.setParticipantIdToCrfEntry(studyId,"STUDY");
   pM.updateCrfEntry();
}

function updateRegistration(){
	let fName="[updateRegistration]";
	print(fName);
	let pM=getParticipantManagerObject(config);
	let idFieldName=pM.getCrfEntryFieldName("STUDY");
	//have to reload query data
	let regQueryPars=parseVariables(config.formConfig.settings['registrationQuery']);
   let regQuery=regQueryPars['query'];
	let fQuery=getQuerySnapshot(regQuery);

	if (fQuery.rows.length==0) {
		print(fName+" registration is empty");
		return; //registration is empty
	}
	let regEntry=fQuery.rows[0];

	for (x in regEntry){
		print(fName+" ["+x+"] "+regEntry[x]);
	}


	let studyId=fQuery.rows[0][idFieldName];
	if (!studyId) {
		print(fName+" study id not set ("+idFieldName+'/'+studyId+")");
		return; //study id not set
	}
	
	//set 
	pM.setParticipantIdToCrfEntry(studyId,"STUDY");
	//this will only update crfEntry in memory, but not on LabKey, 
	//we are counting on updateFlag to follow updateRegistration

	//update parentCRF as well, here we schedule update of data entry as well
	if ("parentCrfData" in config.formConfig){
		let parentCrfEntry=config.formConfig.parentCrfData.rows[0];
		parentCrfEntry[idFieldName]=studyId;
		let cb=function(data){completeWithFlag(data,doNothing);};
      cvModifyRows('update','lists','crfEntry',[parentCrfEntry],cb,getContainer('CRF'));
	}
}

function afterData(){
	let fName='afterData';
	//operatorBasedAccessMode	
	let accessMode=config.formConfig.operator+'Mode';
	let rowsSetup=config.formConfig.formSetupRows;

   let idMode=config.formConfig.form['idMode'];
   //set default value if no value is in the list (read value is null)
   if (!idMode) idMode="STUDY:EDIT";

   print(fName+': idMode '+idMode);
   //add print to config so participantManager can use it
   config.print=print;
   let pM=getParticipantManagerObject(config);
   pM.updateCrfEntry=function(){updateFlag(config.formConfig.crfEntry['FormStatus'],doNothing);};   

   let idModeArray=idMode.split(':');
   pM.mode="STUDY";
   if (idModeArray.includes("LOCAL")) {
      pM.mode="LOCAL";
      //OK, but check if CRF or registration indicate that study id is already set
      verifyCrfStudyId(pM);
	  //study id should already be set by updateRegistration
      //verifyRegistration(pM);
   }
   if (idModeArray.includes("READONLY")){
      pM.readOnly="TRUE";
   }
   
   let pId=pM.getParticipantIdFromCrfEntry();
   if (!pId){
      pM.setEditMode();
   }
   else{
      let label=pId;
      if (pM.mode=="STUDY"){
         let loc=pM.getParticipantIdFromCrfEntry('LOCAL');
         label=pId+':'+loc;
         pM.readOnly="true";
      }
      pM.setLabelMode(label);
      //in STUDY mode also change LOCAL ID from crfEntry
      
   }

	for (let i=0;i<rowsSetup.length;i++){
		let entry=rowsSetup[i];
		let queryName=config.formConfig.queryMap[entry['queryName']];
		
		print(fName+" ["+queryName+"]: showFlag: "+entry["showFlag"]);
		print(fName+" ["+queryName+"]: accessMode: "+entry[accessMode]);

		const nData=getQuerySnapshot(queryName).rows.length;
		
		print(fName+" ["+queryName+"]: nData: "+nData);


		//skip sections
		//also from fields
		if (entry[accessMode]=="NONE") continue;
		//skip readonly empty records
		//if (entry[accessMode]=="READ" && nData==0) continue;
			
		//let additionalData=new Object();
		//setAdditionalData(additionalData,entry);
		//section fits one dataset/list
      
		generateSection(entry);
		//generateSection(queryName,entry["title"],entry[accessMode], 
		//		additionalData);
	}

}

function findSetupRow(queryName,formId){
	let rowsSetup=config.formConfig.formSetupRows;
	for (let i=0;i<rowsSetup.length;i++){
		let e=rowsSetup[i];
		let queryName1=config.formConfig.queryMap[e['queryName']];
		if (e.formName!=formId) continue;
		if (queryName1!=queryName) continue;
		return e;
	}
	return null;
}

function populateSection(queryName){
	let fName='[populateSection/'+queryName+']';
	print(fName);

   //old setting
   let formId=config.formId;
   //new setting
   formId=config.formConfig.formId;

	let entry=findSetupRow(queryName,formId);
	//ignore names without associated entry in formSetup
	if (entry==undefined){
		print(fName+': no matching FormSetup entry found');
		return;
	}
	//populate comes after generate, we should be pretty safe in taking
	//already generated additionalData
	
	if (!(queryName in config.formConfig.additionalData)){
		print(fName+': no additionalData generated for '+queryName);
		return;
	}
	
	let additionalData=config.formConfig.additionalData[queryName];
	print(fName+': using additionalData '+additionalData);
	if ("isReview" in additionalData){
		generateReviewSection(queryName,queryName,generateReviewSectionCB);
		return;	
	}

	let accessMode=config.formConfig.operator+'Mode';
	let aM=entry[accessMode];
	print(fName+': accessMode '+aM);

	if (aM!='GENERATE'){
		let writeMode=entry[accessMode]=='EDIT';
		print(fName+': mode='+writeMode);
		populateTable(queryName,writeMode);
		return;
	}

	//deal with generate
	//
	//already available -> shift to READ mode
	let divTable=queryName+'Table';
	let divObj=config.document.getElementById(divTable);
	let divRev=config.document.getElementById(queryName+'Review');
	let divRLi=config.document.getElementById(queryName+'ReviewList');
	let divGBu=config.document.getElementById(queryName+'GenerateButton');

	print('div GBU: '+divGBu);
	divObj.style.display="block";
	divRev.style.display="block";
	divRLi.style.display="block";
	if (divGBu!=undefined) divGBu.style.display="none";

	let nData=getQuerySnapshot(queryName).rows.length;
	print('['+queryName+']: nrows '+nData);
	if (nData>0){
		populateTable(queryName,0);
		return;
	}
	//hide table
	divObj.style.display="none";
	divRev.style.display="none";
	divRLi.style.display="none";
	if (divGBu!=undefined) divGBu.style.display="block";
	//add buttons?
	//is button already generated?
	
	//populateTable(entry);
	
}		

//*******    generateQuery infrastructure *********************

function onGenerateQuery(queryName){
   let fName='[onGenerateQuery]';
	print(fName+' '+queryName);
//
	let cfgRows=config.formConfig.generateConfigData.rows;
//	//queryName to queryId?
	let queryId=config.formConfig.fields[queryName].queryId;
	let cfgRow=undefined;

	for (let i=0;i<cfgRows.length;i++){
		if (cfgRows[i].queryId!=queryId) continue;
		cfgRow=cfgRows[i];
		break;
	}
	if (cfgRow==undefined){
		print('generateConfig for queryName['+queryId+']='+queryName+' not found');
		return;
	}
	//add config to the list
	if (!("generateConfig" in config.formConfig)){
		config.formConfig.generateConfig=new Object();
	}
	config.formConfig.generateConfig[queryName]=cfgRow;

	if (!("generateForm" in config.formConfig)){
		config.formConfig.generateForm=new Object();
	}
	
//
	let formRows=config.formConfig.formRows;	
	let formId=cfgRow.formId;
	for (let i=0;i<formRows.length;i++){
		if (formRows[i].Key==formId) {
			config.formConfig.generateForm[queryName]=formRows[i];
			break;
		}
	}
	//print('XcfgRow '+config.formConfig.generateForm[queryName]);
//
//	//check if all required datasets were at least saved
	checkGenerationFields(queryName);
}

function checkGenerationFields(queryName){
   let fName='[checkGenerationFields]';
	let genForm=config.formConfig.generateForm[queryName];
	let genCfg=config.formConfig.generateConfig[queryName];
	let mailRecipient=genCfg.emailRecipient;
	
	//list of queries that are part of Registration form
	print(fName);	
	print(fName+' setRecipient: '+mailRecipient);
	let formId=genForm.Key;
	print(fName+" Checking form w/id "+formId);
	let selectGenerationRows=selectFormSetupRows(formId);
	//registration rows
	for (let i=0;i<selectGenerationRows.length;i++){
		let row=selectGenerationRows[i];
		let queryId=row.queryName;
		let fQueryName=config.formConfig.queryMap[queryId];
		if (fQueryName==queryName) continue;
		let fQuery=getQuerySnapshot(fQueryName);
		print('Checking '+fQueryName+' nrows: '+fQuery.rows.length);
		if (fQuery.rows.length==0){ 
			generateError(queryName,fQueryName);
			return;
		}
	}
	generateMessage(queryName,'Vailidation OK');
	print('callback: set recipient: '+mailRecipient);
	let cb=function(){prepareForm(queryName,formId,mailRecipient);};
	generateListEntry(formId,queryName,cb);
}



function prepareForm(queryName,formId,mailRecipient){
   let fName="[prepareForm]";

	print(fName+' recipient '+mailRecipient);

	//look for existing registration entry
	let action=function(data){generateForm(data,queryName,mailRecipient);};
	let formFilter=LABKEY.Filter.create('Form',formId);
	let parentCrfFilter=LABKEY.Filter.create('parentCrf',getCRFref());
   let filters=[formFilter,parentCrfFilter];
   cvSelectRows('lists','crfEntry',filters,action,getContainer('data'));

}

function generateError(queryName,fQueryName){
	let elName=queryName+'GenerateButton'+'_reportField';
	let el=config.document.getElementById(elName);
	el.innerText='Error: '+fQueryName+' was not set';
	el.style.color='red';
}
function generateMessage(queryName,msg){
	let elName=queryName+'GenerateButton'+'_reportField';
	let el=config.document.getElementById(elName);
	el.innerText=msg;
	el.style.color='green';
}

function generateForm(data,queryName,mailRecipient){

   let fName='[generateForm]';

	print(fName+' recipient: '+mailRecipient);
//	
	const nData=data.rows.length;
	print(fName+' Registration: '+nData+' rows');

	let formRow=config.formConfig.generateForm[queryName];
	let formCfg=config.formConfig.generateConfig[queryName];

	//we have to generate masterQuery with parentCrf and crfRef 
	//and crfEntry with new entryId and parentCrf equal to crfRef
	if (nData>0) {
		generateMessage(queryName,'Registration already generated.');
		return;
	}
	let formId=formRow.Key;
	let formName=formRow.formName;
	let crfBase=config.formConfig.crfEntry;
	let crfEntry=new Object();
	//add new reference
	crfEntry.entryId=Date.now();
	crfEntry.parentCrf=getCRFref();
	crfEntry["Date"]=new Date();
	crfEntry["View"]="[VIEW]";

	crfEntry.formStatus=1;//In progress
   //checks for both field presence (if not in query, undefined) and field value (if not set, null)
   print(fName+' setup status: '+formCfg.formStatus);
   if (formCfg.formStatus){
      crfEntry.formStatus=formCfg.formStatus;
   }

   //get local Id
   let pM=getParticipantManagerObject(config);
   
   crfEntry[pM.getCrfEntryFieldName()]=pM.getParticipantIdFromCrfEntry();
//	//set other variables
	//requires studyData as part of formConfig
//	let studyData=config.formConfig.studyData;
	print('Adding study: '+crfBase.EudraCTNumber);
	crfEntry.EudraCTNumber=crfBase.EudraCTNumber;
	crfEntry.StudyCoordinator=crfBase.StudyCoordinator;
	crfEntry.StudySponsor=crfBase.StudySponsor;
	crfEntry.RegulatoryNumber=crfBase.RegulatoryNumber;
//
//	//find sponsor for site
	let site=crfBase.Site;
	let crfSponsors=config.formConfig.crfSponsors;
	let users=config.formConfig.userRows;
	for (let i=0;i<crfSponsors.length;i++){
		//print('Checking for site '+crfSponsors[i].Site);
		if (crfSponsors[i].Site!=site) continue;
		config.formConfig.sponsorId=crfSponsors[i].User;
		//print('Setting id '+config.formConfig.sponsorId);
		//finds first
		break;
	}
	for (let j=0;j<users.length;j++){
		if (config.formConfig.sponsorId!=users[j].UserId) continue;
		config.formConfig.sponsor=users[j];
		//finds first (should be unique)
		break;
	}
	print('Selecting '+config.formConfig.sponsor.DisplayName+' as sponsor');
	//different user than the original form...
	//should be set to the study sponsor
	crfEntry.UserId=config.formConfig.sponsor.UserId;
	crfEntry.Site=site;
//	//set formId to one found through registration search
	crfEntry.Form=formId;
////

   let crfStatus=createCrfStatus(crfEntry);
   crfStatus.operator=config.formConfig.operator;
   crfStatus.action='generateForm';


	let cb=function(data){sendEmail(data,mailRecipient,doNothing,formName+' generated');}
   let pass=function(data){cvInsertRows('lists','crfStatus',[crfStatus],cb,getContainer('data'));};
   cvInsertRows('lists','crfEntry',[crfEntry],pass,getContainer('data'));

}
//
function generateListEntry(formId,queryName,cb){

	//check if registration was already generated

	let formRows=config.formConfig.formRows;
	let qForm=undefined;
	for (let i=0;i<formRows.length;i++){
		if (formRows[i].Key!=formId) continue;
		qForm=formRows[i];
	}
	let nData=getQuerySnapshot()[queryName].rows.length;

   if (nData>0) return;


   //create new list entry
   let pM=getParticipantManagerObject(config);
   

	let e2=new Object();
	e2.crfRef=getCRFref();
	e2.registrationStatus=0;
	e2.submissionDate=new Date();
   e2[pM.getCrfEntryFieldName()]=pM.getParticipantIdFromCrfEntry();
	print('set values');

   cvInsertRows('lists',queryName,[e2],cb,getContainer('data'));

}
		
// ******************** end form generator (Registration) ********************

//jump to populate table/generate review, etc defined at the begining of the file

function setContainer(label,container){
	if (!(config.formConfig.hasOwnProperty('container'))){
		config.formConfig.container=new Array();
	}
	config.formConfig.container[label]=container;
}

function getContainer(label){
	return config.formConfig.container[label];
}

//entry point from generateMasterForm
function setFormConfig(){

	

	//add object to store form related data
	config.formConfig=new Object();

	config.formConfig.softwareVersion='T.15.68';
	let debug=true;

	if (debug)
		print("generateMasterForm1");	
	
	//set containers for data and configuration

	//TODO: set this from a query
	//
	
	setContainer('data',LABKEY.ActionURL.getContainer());
	setContainer('config',LABKEY.ActionURL.getContainer());
	setContainer('CRF',LABKEY.ActionURL.getContainer());

	//this is local data
   cvSelectRows('lists','crfSettings',[],afterSettings,getContainer('CRF'));
	//store form related data to this object

}


function convertToDictionary(rows){
   let x=new Array();
	for (let i=0;i<rows.length;i++){
		let n=rows[i]['name'];
		let v=rows[i]['value'];
		x[n]=v;
	}
   return x;
}
   
function convertToAssociatedArray(rows,fieldName="name"){
   let fName="[convertToAssociatedArray]";
   let x=new Object();
	for (let i=0;i<rows.length;i++){
		let n=rows[i][fieldName];
		x[n]=rows[i];
	}
   return x;
}
 
function afterSettings(data){

	config.formConfig.settings=convertToDictionary(data.rows);

	let st=config.formConfig.settings;
	print('afterSettings');
	for (let k in st){
		print('\t'+k+'='+st[k]);
	}

	//if ('dataContainer' in st){
	//	setContainer('data',st['dataContainer']);
	//}
	let vname='configContainer';
	if (vname in st){
		setContainer('config',st[vname]);
	}
	print('Config: '+getContainer('config'));
	print('Data: '+getContainer('data'));

	//use first-> we must first establish link to the rigth crf entry
	let filters=[LABKEY.Filter.create('entryId',getCRFrefFirst())];
   cvSelectRows('lists','crfEntry',filters,afterCRFEntry,getContainer('data'));
}

function afterCRFEntry(data){
	config.formConfig.crfEntry=data.rows[0];
	print("Setting crfEntry (x) to "+config.formConfig.crfEntry["entryId"]);
	//for empty records or those with parentCrf not set, parentCrf comes up as null
	//nevertheless, with two equal signs, check against undefined also works
	print('parentCrf set to '+config.formConfig.crfEntry.parentCrf);
	
	collectData();
}

function collectData(){
	
   let targetObject=config.formConfig;
   targetObject.print=print;
   targetObject.getContainer=getContainer;
	let queryArray=new Array();
   //k
	//site
	queryArray.push(makeQuery(targetObject,'config','site','siteData',[]));
	//users
	queryArray.push(makeQuery(targetObject,'CRF','users','userData',[]));
	queryArray[queryArray.length-1].schemaName='core';
	//crfEditors
	queryArray.push(makeQuery(targetObject,'config','crfEditors','crfEditorsData',[]));
	//crfMonitors
	queryArray.push(makeQuery(targetObject,'config','crfMonitors','crfMonitorsData',[]));
	//crfSponsors
	queryArray.push(makeQuery(targetObject,'config','crfSponsors','crfSponsorsData',[]));

   //study static data
   queryArray.push(
      makeQuery(targetObject,'data','crfStaticVariables','crfStaticVariables',[]));

   queryArray.push(makeQuery(targetObject,'data','specialFields','specialFieldsQuery',[]));
	//study
	queryArray.push(makeQuery(targetObject,'data','Study','studyDataAll1',[]));
	let e=queryArray[queryArray.length-1];
	//overload schema name
	e.schemaName='study';
	//make sure variables not part of default view are loaded
	//here we should already have read crfStaticVariables table
	e.columns="SubjectColumnName,EudraCTNumber,StudySponsor";
	e.columns+=",StudyCoordinator,RegulatoryNumber";
	
	//formStatus
	let varLabel='sourceFormStatus';
	let formStatus=config.formConfig.crfEntry['FormStatus'];
	let formFilter=LABKEY.Filter.create('Key',formStatus);
	queryArray.push(
		makeQuery(targetObject,'config','FormStatus','formStatusData',[formFilter]));
	//crfButtons
	let statusFilter=LABKEY.Filter.create(varLabel,formStatus);
	queryArray.push(
		makeQuery(targetObject,'config','crfButtons','crfButtons',[statusFilter]));
	//Forms	
	queryArray.push(makeQuery(targetObject,'config','Forms','formData',[]));
	//FormSetup	
	queryArray.push(makeQuery(targetObject,'config','FormSetup','formSetup',[]));
	//generateConfig
	queryArray.push(
		makeQuery(targetObject,'config','generateConfig','generateConfigData',[]));	
	
   //inputLists 
	queryArray.push(
		makeQuery(targetObject,'config','inputLists','inputLists',[]));	
   //parentCrf
	let parentCrf=config.formConfig.crfEntry['parentCrf'];
	if (parentCrf!=undefined){
		let crfFilter=LABKEY.Filter.create('entryId',parentCrf);
		queryArray.push(makeQuery(targetObject,'data','crfEntry','parentCrfData',[crfFilter]));	
	}	

	print('running getDataFromQueries');
	getDataFromQueries(queryArray,addStudyData);
}

function addStudyData(){
   let fName='addStudyData';

   //convert specialFields to array
   let q=config.formConfig["specialFieldsQuery"].rows;
   config.formConfig.specialFields=convertToAssociatedArray(q,"fieldUID");
   print(fName);
	let queryArray=new Array();
	
   let targetObject=config.formConfig;
   targetObject.print=print;
   targetObject.getContainer=getContainer;
	//study
	queryArray.push(makeQuery(targetObject,'data','Study','studyDataAll',[]));
	//queryArray.push(makeQuery('data','Study','studyDataAll',[]));
	let e=queryArray[queryArray.length-1];
	//overload schema name
	e.schemaName='study';
	//make sure variables not part of default view are loaded
	//here we should already have read crfStaticVariables table
   let staticVarRows=config.formConfig['crfStaticVariables'].rows;
   let columnModel=""
   for (let i=0;i<staticVarRows.length;i++){
      if (i>0) columnModel+=',';
      columnModel+=staticVarRows[i]['staticVariable'];
   }
	e.columns=columnModel;

   //also collect ids already in study
   //registrationQuery should be a dataset
   //since monitors can review late, it might be profitable to use lists
   //rather than study
   let regQueryPars=parseVariables(config.formConfig.settings['registrationQuery']);
   let regQuery=regQueryPars['query'];
   let regSchema='study';
   if ('schema' in regQueryPars){
      regSchema=regQueryPars['schema'];
   }
   queryArray.push(makeQuery(targetObject,'data',regQuery,'registrationData',[]));
   queryArray[queryArray.length-1].schemaName=regSchema;
      
	
	getDataFromQueries(queryArray,fcontinue);

}

function selectFormSetupRows(formId){
	let formSetupRows=new Array();
	let allRows=config.formConfig.formSetup.rows;
	for (let i=0;i<allRows.length;i++){
		let formEntry=allRows[i];
		if (formEntry.formName==formId)
			formSetupRows.push(formEntry);
	}
	return formSetupRows;
}

function fcontinue(){

   //debug
   let fName='[fcontinue]';
   let varRows=config.formConfig['crfStaticVariables'].rows;
   let studyVars=config.formConfig['studyDataAll'].rows[0];
   for (let i=0;i<varRows.length;i++){
      let vName=varRows[i].staticVariable;
      print(fName+' '+vName+': '+studyVars[vName]);
   }

	//parse site
	config.formConfig.siteRows=config.formConfig.siteData.rows;
	let sRows=config.formConfig.siteRows;
	for (let i=0;i<sRows.length;i++){
		let siteId=sRows[i].siteNumber;
		print('site '+siteId);
		if (siteId==config.formConfig.crfEntry.Site){
			config.formConfig.currentSite=sRows[i];
			break;
		}
	}
	//config.formConfig.site=data.rows[0];
	print("Setting site name to "+config.formConfig.currentSite.siteName);
	//study
	config.formConfig.studyData=config.formConfig.studyDataAll.rows[0];
	print("XSetting participantField to "+
			config.formConfig.studyData["SubjectColumnName"]);
	
	config.formConfig.crfEditors=config.formConfig.crfEditorsData.rows;
	config.formConfig.crfMonitors=config.formConfig.crfMonitorsData.rows;
	config.formConfig.crfSponsors=config.formConfig.crfSponsorsData.rows;

	config.formConfig.userRows=config.formConfig.userData.rows;
	let uRows=config.formConfig.userRows;
	for (let i=0;i<uRows.length;i++){
		let userId=uRows[i].UserId;
		if (userId==config.formConfig.crfEntry.UserId){
			config.formConfig.user=uRows[i];
			break;
		}
	}
	//config.formConfig.user=data.rows[0];
	print("Setting user to "+config.formConfig.user["DisplayName"]);



	let fsRows=config.formConfig.formStatusData.rows;
	config.formConfig.formStatus=fsRows[0].formStatus;
	config.formConfig.operator=config.role;
	//config.formConfig.operator=fsRows[0].operator;

	print('Setting operator to: '+config.formConfig.operator);
	
	config.formConfig.formRows=config.formConfig.formData.rows;

   //point formId to point to form set in crfEntry
   config.formConfig.formId=config.formConfig.crfEntry['Form'];

   //old setting, set from URL in visit.html
   let formId=config.formId;
   //new setting, set from crfEntry
   formId=config.formConfig.formId;

	let formRows=config.formConfig.formRows;
	//filter out the current form
	for (let i=0;i<formRows.length;i++){
		if (formRows[i].Key==formId){
			config.formConfig.form=formRows[i];
			break;
		}
	}
	
	config.formConfig.formSetupRows=selectFormSetupRows(formId);

	print("Number of datasets for form ["+formId+"]: "+
			config.formConfig.formSetupRows.length);



	let fields=config.formConfig.formSetup.metaData.fields;

	//get the lookup for queryName column
	let formQueryName='queryName';
	let field="NONE";
	for (f in fields){
		if (fields[f]['name']!=formQueryName) continue;
		field=fields[f];
		break;
	}
	let lookup=field.lookup;

	print("Getting dataset names from "+lookup.queryName);

	//inputLists should be in configuration container
   cvSelectRows(lookup.schemaName,lookup.queryName,[],afterFormDatasets,getContainer('config'));

}

function afterFormDatasets(data){
	print('afterFormDatasets: '+data.rows.length);
	config.formConfig.formDatasets=data;//inputLists
	config.formConfig.fields=new Object();
	config.formConfig.queryMap=new Object();
	config.formConfig.additionalData=new Object();

	let rows=config.formConfig.formSetupRows;

	//should skip report only rows
	for (let i=0;i<rows.length;i++){
		let entry=rows[i];
		let reviewField=(entry['showFlag']=='REVIEW');
		//is the operator set yet?
		let accessMode=config.formConfig.operator+'Mode';
		let skipField=(entry[accessMode]=="NONE");
		let queryId=entry['queryName'];
		let lookupRows=config.formConfig.formDatasets.rows;
		print('QueryID['+i+']='+queryId);
		let dentry;

		for (let j=0;j<lookupRows.length;j++){
	
			if (queryId!=lookupRows[j]['Key']) continue;
			dentry=lookupRows[j];
			break;
		}
		let qName=dentry['queryName'];

		//update list of dataset formConfig is observing (fields/queryMap)
		while (1){
			//review contains no data
			if (reviewField) break;
			if (skipField) break;
			//already in fields
			if (qName in config.formConfig.fields) break;
			config.formConfig.fields[qName]=new Object();
			break;
		}

		while(1){
			//already done
			if (queryId in config.formConfig.queryMap) break;
			config.formConfig.queryMap[queryId]=qName;
			break;
		}
		if (reviewField) continue;
		if (skipField) continue;
		//only do this for real lists
		let field=config.formConfig.fields[qName];
		field.title=entry['title'];
		field.queryId=queryId;

	}
	print("List of datasets in form : ");
	for (f in config.formConfig.fields){
		let field=config.formConfig.fields[f];
		print("\t"+f+" ID: "+field.queryId+' title '+field.title);
	}
	afterConfig();

}

//>>>>>>>>>>>>>>>>>new>>>>>>>>>>>>

function setDataLayout(cb){
   let fName='[setDataLayout]';
   print(fName);
	let rowsSetup=config.formConfig.formSetupRows;
   let queryArray=new Array();
	let dS=getLayoutObject();//reference only
   let qList=getQueryList();
	let qMap=config.formConfig.queryMap;
	//config.formConfig.lookup=new Object();
	for (let i=0;i<rowsSetup.length;i++){
		let entry=rowsSetup[i];
		//skip review rows
		if (entry['showFlag']=='REVIEW')
			continue;
		let queryId=entry['queryName'];
		let q=qMap[queryId];
		queryArray.push(makeQuery(dS,'data',q,q,[]));
      qList[q]=0;
      print(fName+' adding '+q);
		if (entry['showQuery']!="NONE"){
			let sq=entry['showQuery'];
		   queryArray.push(makeQuery(dS,'data',sq,sq,[]));
         qList[sq]=0;
         print(fName+' adding '+sq);
			
		}
	}

	//always add reviews
   let q='reviewComments';
   queryArray.push(makeQuery(dS,'data',q,q,[]));
   qList[q]=0;
   let action=function(){processLayout(cb);};
   getDataFromQueries(queryArray,action);
}

function findTitle(queryName){
//find by name from formDatasets 
//and set associated title as title
	let rows=config.formConfig.formDatasets.rows;
	for (let i=0;i<rows.length;i++){
		let entry=rows[i];
		if (entry['queryName']!=queryName) continue;
		return entry['title'];
	}
	return "NONE";
}

//this happens after the for loop, so all dataQueries objects are set
function processLayout(cb){
   let fName='[processLayout]';
   let qList=getQueryList();
   //for layouts
   let queryArray=new Array();
   let targetObject=getLookupObject();
   let lookupSet=new Object();
   for (let q in qList){
      let qobject=getQueryLayout(q);
	   print(fName+" inspecting layout for "+q+" "+qobject);
	   qobject.fields=qobject.metaData.fields;
	   qobject.title=findTitle(q);

      //check for lookups
	   for (let f in qobject.fields){
		   //anything else is simple but lookup
		   let field=qobject.fields[f];
		   if (!("lookup" in field)) continue;
         let lookup=field.lookup;
         let qObject=getLookup(lookup.queryName);
         if (qObject) continue;
         //add to list
         let qName=lookup.queryName;
         let qCode=qName+':'+lookup.keyColumn+':'+lookup.displayColumn;
         let e=makeQuery(targetObject,'data',qName,qCode,[]);
         //adjust minor settings
         if (lookup.containerPath) e.containerPath=lookup.containerPath;
         e.schemaName=lookup.schemaName;
         e.columns=lookup.keyColumn+','+lookup.displayColumn;
         lookupSet[qCode]=e;
         print(fName+' inserting '+qCode);
      }
   }
   for (let x in lookupSet){
      queryArray.push(lookupSet[x]);
      print(fName+' adding '+x);
      for (let v in lookupSet[x]){
         print(fName+' value ['+v+'] '+lookupSet[x][v]);
      }
   }
   //print(fName+' print '+targetObject.print);
   let action=function(){processLookup(cb);};
   print(fName+' getDataFromQueries');
   getDataFromQueries(queryArray,action);
   print(fName+' getDataFromQueries done');
}

function processLookup(cb){
   let fName="[processLookup]";
   let obj=getLookupObject();
   for (let q in obj){
	   print(fName+" "+q);
      let a=q.split(':');
      if (a.length<3) continue;
      let lookupName=a[0];
      let key=a[1];
      let val=a[2];
      obj[lookupName]=new Object();
      print(fName+' adding ['+lookupName+'] '+key+'/'+val);
      let lObject=obj[lookupName];

	   lObject.LUT=new Array();//key to value
	   lObject.ValToKey=new Array();//value to key
	   lObject.keyColumn=key
	   lObject.displayColumn=val;
      
      let qRows=obj[q].rows;
	   for (let i=0;i<qRows.length;i++){
         let r=qRows[i];
         print(fName+' LUT ['+r[key]+'] '+r[val]);
		   lObject.LUT[r[key]]=r[val];
		   lObject.ValToKey[r[val]]=r[key];
	   }
   }
	cb();
}

function setData(cb){
	fName='[setData]';
	let crfMatch=getCRFref();
	let parentCrf=config.formConfig.crfEntry['parentCrf'];
	if (parentCrf!=undefined) crfMatch=parentCrf;

	print(fName+' form crf ['+getCRFref()+'] matching for crfRef='+crfMatch);

   let queryArray=new Array();
   let targetObject=getSnapshotObject();
	//collect data and execute callback cb for queries in cb.queryList
   let qList=getQueryList();
	for (q in qList){

		let filters=[LABKEY.Filter.create("crfRef",crfMatch)];
      queryArray.push(makeQuery(targetObject,'data',q,q,filters));

	}
   getDataFromQueries(queryArray,cb);
}

function uploadFile(inputElement,context){
	//context should have ID and dirName attributes; 
	//path will be dirName/ID/fieldName_ID.suf
	//where suf is identical to localPath content picked from
	//inputElement
	print('uploadFile: '+inputElement.value+'/');
	if (inputElement.type=="text") return;
	print('uploadFile: '+inputElement.files+'/');
	print('uploadFile: '+inputElement.files.length+'/');
	if (inputElement.files.length>0){
		let file=inputElement.files[0];
		print('uploadFile: '+inputElement.value+'/'+file.size);
	}

	let url=LABKEY.ActionURL.getBaseURL();
	url+='_webdav';
	url+=LABKEY.ActionURL.getContainer();
	url+='/@files';
	url+='/'+context['dirName'];

	print('uploadFile url: '+url);
	let uploadConfig=new Object();
	uploadConfig.inputElement=inputElement;
	uploadConfig.context=context;
	uploadConfig.url=url;
	uploadConfig.success=afterBaseDir;
	uploadConfig.failure=tryMakeDir;
	webdavCheck(uploadConfig);
}

function afterBaseDir(cfg){
	print('afterBaseDir');
	cfg.url+='/'+cfg.context['ID'];
	cfg.success=afterIDDir;
	cfg.failure=tryMakeDir;
	webdavCheck(cfg);
}

function afterIDDir(cfg){
	print('afterIDDir');
	if (cfg.inputElement.files.length==0){
	       	print('No files found');	
		return;
	}
	let file=cfg.inputElement.files[0];
	print('Uploading '+file.name);
	let suf=file.name.split('.').pop();
	cfg.url+='/'+cfg.context['ID']+'.'+suf;
	cfg.success=afterUpload;
	cfg.failure=onFailure;
	cfg.data=file;
	webdavPut(cfg);
}

function afterUpload(cfg){
	print('afterUpload');
}

function tryMakeDir(cfg){
	print('tryMakeDir '+cfg.url);
	cfg.failure=onFailure;
	webdavMakeDir(cfg);
}





function request(cfg,verb,data){
	print('request['+verb+'] '+cfg.url);
	let connRequest=new XMLHttpRequest();
	connRequest.addEventListener("loadend",
		function(){checkResponse(connRequest,cfg);});
	connRequest.open(verb, cfg.url);
	connRequest.send(data);
	//print('request['+verb+'] sent');
}

function checkResponse(xrq,cfg){
	//print('checkResponse: readyState '+xrq.readyState);
	//print('checkResponse: status '+xrq.status);
	if (xrq.status<400) {
		//client errors 400-499
		//server errors 500-599
		cfg.success(cfg);
		return;
	}
	cfg.status=xrq.status;
	cfg.failure(cfg);
}

function webdavMakeDir(cfg){ request(cfg,'MKCOL',null);}
function webdavCheck(cfg) { request(cfg,'GET',null);}
function webdavPut(cfg) { request(cfg,'PUT',cfg.data);}


function onFailure(cfg){
	print('request failed with status='+cfg.status);
}

